(*
 * Define an extendable bit-vector.
 * This is an imperative data structure.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2004 Mojave Group, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

(*
 * Bit vectors are represented with strings.
 * The vector is represented least-significant byte first.
 *)
type t = string ref

(*
 * Byte size.
 *)
let byte_shift = 3
let byte_mask  = (1 lsl byte_shift) - 1

(*
 * Empty vector.
 *)
let create () =
   ref ""

(*
 * Actual length of a vector.
 *)
let length buf =
   String.length buf lsl byte_shift

(*
 * Grow the vector.
 *)
let extend vec i =
   let buf1 = !vec in
   let len = length buf1 in
      if i >= len then
         let buf2 = String.make ((i lsr byte_shift) + 1) '\000' in
            String.blit buf1 0 buf2 0 (String.length buf1);
            vec := buf2

(*
 * Add a bit.
 *)
let set vec i =
   extend vec i;
   let buf = !vec in
   let j = i lsr byte_shift in
   let k = i land byte_mask in
      buf.[j] <- Char.chr ((Char.code buf.[j]) lor (1 lsl k))

(*
 * Probe a bit.
 *)
let test vec i =
   let buf = !vec in
   let len = length buf in
   let j = i lsr byte_shift in
   let k = i land byte_mask in
      i < len && ((Char.code buf.[j]) land (1 lsl k)) != 0

(*
 * Add bits in the second vector to the first.
 *)
let union vec1 vec2 =
   let buf1 = !vec1 in
   let buf2 = !vec2 in
   let len1 = String.length buf1 in
   let len2 = String.length buf2 in
   let buf1 =
      if len2 > len1 then
         let buf1' = String.make len2 '\000' in
            String.blit buf1 0 buf1' 0 len1;
            vec1 := buf1';
            buf1'
      else
         buf1
   in
      for i = 0 to pred len2 do
         buf1.[i] <- Char.chr (Char.code buf1.[i] lor Char.code buf2.[i])
      done

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
